from __future__ import annotations

import contextlib
import enum
from functools import cached_property
from typing import Any, Generic, Optional, Type, TypeVar

from attrs import define

from qtgql.codegen.py.compiler.builtin_scalars import BuiltinScalar
from qtgql.codegen.py.runtime.bases import QGraphQListModel
from qtgql.codegen.py.runtime.custom_scalars import BaseCustomScalar, CustomScalarMap
from qtgql.codegen.utils import AntiForwardRef
from qtgql.utils.typingref import UNSET, TypeHinter


class Kinds(enum.Enum):
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    ENUM = "ENUM"
    LIST = "LIST"
    UNION = "UNION"
    NON_NULL = "NON_NULL"
    INTERFACE = "INTERFACE"
    INPUT_OBJECT = "INPUT_OBJECT"


@define(slots=False)
class QtGqlBaseTypedNode:
    name: str
    type: GqlTypeHinter
    type_map: dict[str, QtGqlObjectTypeDefinition]
    enums: EnumMap
    scalars: CustomScalarMap

    @property
    def is_custom_scalar(self) -> Optional[Type[BaseCustomScalar]]:
        return self.type.is_custom_scalar(self.scalars)

    @cached_property
    def annotation(self) -> str:
        """
        :returns: Annotation of the field based on the real type,
        meaning that the private attribute would be of that type.
        this goes for init and the property setter.
        """
        return self.type.annotation(self.scalars)


T = TypeVar("T")


@define(slots=False)
class QtGqlVariableDefinition(Generic[T], QtGqlBaseTypedNode):
    default_value: Optional[T] = UNSET

    def json_repr(self, attr_name: Optional[str] = None) -> str:
        if not attr_name:
            attr_name = self.name
        if self.type.is_input_object_type:
            return f"{attr_name}.asdict()"
        elif self.type.is_builtin_scalar:
            return attr_name
        elif self.type.is_enum:
            return f"{attr_name}.name"
        elif self.is_custom_scalar:
            return f"{attr_name}.{BaseCustomScalar.parse_value.__name__}()"

        raise NotImplementedError(f"{self.type} is not supported as an input type ATM")


@define(slots=False)
class BaseQtGqlFieldDefinition(QtGqlBaseTypedNode):
    description: Optional[str] = ""


@define(slots=False)
class QtGqlInputFieldDefinition(BaseQtGqlFieldDefinition, QtGqlVariableDefinition):
    ...


@define(slots=False)
class QtGqlFieldDefinition(BaseQtGqlFieldDefinition):
    @cached_property
    def default_value(self):
        if builtin_scalar := self.type.is_builtin_scalar:
            if builtin_scalar.tp is str:
                return f"'{builtin_scalar.default_value}'"
            return f"{builtin_scalar.default_value}"
        if self.type.is_object_type:
            return "None"

        if self.type.is_model:
            # this would just generate the model without data.
            return "list()"

        if custom_scalar := self.type.is_custom_scalar(self.scalars):
            return f"SCALARS.{custom_scalar.__name__}()"

        if enum_def := self.type.is_enum:
            return f"{enum_def.name}(1)"  # 1 is where auto() starts.

        return "None"  # Unions are not supported yet.

    @cached_property
    def fget_annotation(self) -> str:
        """This annotates the value that is QML-compatible."""
        if custom_scalar := self.type.is_custom_scalar(self.scalars):
            return TypeHinter.from_annotations(
                custom_scalar.to_qt.__annotations__["return"],
            ).stringify()
        if self.type.is_enum:
            return "int"

        return self.annotation

    @cached_property
    def property_type(self) -> str:
        try:
            # this should raise if it is an inner type.
            ret = GqlTypeHinter.from_string(self.fget_annotation, self.type_map)
            if ret.type in self.type_map.values():
                raise TypeError  # in py3.11 the get_type_hints won't explicitly raise.
            return self.fget_annotation
        except (TypeError, NameError):
            if self.type.is_union():
                # graphql doesn't support scalars in Unions ATM. (what about Enums in unions)?
                return "QObject"
            if self.type.is_enum:
                # QEnum value must be int
                return "int"
            # might be a model, which is also QObject
            assert self.type.is_model or self.type.is_object_type
            return "QObject"

    @cached_property
    def setter_name(self) -> str:
        return self.name + "_setter"

    @cached_property
    def signal_name(self) -> str:
        return self.name + "Changed"

    @cached_property
    def private_name(self) -> str:
        return "_" + self.name

    @property
    def fget(self) -> str:
        if self.type.is_custom_scalar(self.scalars):
            return f"return self.{self.private_name}.{BaseCustomScalar.to_qt.__name__}()"
        if self.type.is_enum:
            return f"return self.{self.private_name}.value"
        return f"return self.{self.private_name}"

    @cached_property
    def can_select_id(self) -> Optional[QtGqlFieldDefinition]:
        object_type = self.type.is_object_type
        if not object_type:
            if self.type.is_model:
                object_type = self.type.is_model.is_object_type
        if object_type:
            return object_type.has_id_field


@define(slots=False)
class BaseGqlTypeDefinition:
    name: str
    fields_dict: dict[str, QtGqlFieldDefinition]
    docstring: Optional[str] = ""

    @property
    def fields(self) -> list[QtGqlFieldDefinition]:
        return list(self.fields_dict.values())


@define(slots=False)
class QtGqlObjectTypeDefinition(BaseGqlTypeDefinition):
    @cached_property
    def has_id_field(self) -> Optional[QtGqlFieldDefinition]:
        return self.fields_dict.get("id", None)

    @cached_property
    def id_is_optional(self) -> Optional[QtGqlFieldDefinition]:
        if id_f := self.has_id_field:
            if id_f.type.is_optional():
                return id_f


@define(slots=False)
class QtGqlInputObjectTypeDefinition(BaseGqlTypeDefinition):
    fields_dict: dict[str, QtGqlInputFieldDefinition] = {}  # type: ignore


@define
class EnumValue:
    """encapsulates enumValues from introspection, maps to an Enum member."""

    name: str
    description: str = ""


@define
class QtGqlEnumDefinition:
    name: str
    members: list[EnumValue]


EnumMap = dict[str, "QtGqlEnumDefinition"]


def optional_maybe(th: GqlTypeHinter) -> GqlTypeHinter:
    return th if not th.is_optional() else th.of_type[0]


class GqlTypeHinter(TypeHinter):
    def __init__(
        self,
        type: Any,
        of_type: tuple[GqlTypeHinter, ...] = (),
    ):
        self.type = type
        self.of_type: tuple[GqlTypeHinter, ...] = of_type

    @property
    def is_object_type(self) -> Optional[QtGqlObjectTypeDefinition]:
        t_self = optional_maybe(self).type
        with contextlib.suppress(TypeError):
            if issubclass(t_self, AntiForwardRef):
                ret = t_self.resolve()
                if isinstance(ret, QtGqlObjectTypeDefinition):
                    return ret

    @property
    def is_input_object_type(self) -> Optional[QtGqlInputObjectTypeDefinition]:
        t_self = optional_maybe(self).type
        if isinstance(t_self, QtGqlInputObjectTypeDefinition):
            return t_self

    @property
    def is_model(self) -> Optional[GqlTypeHinter]:
        t_self = optional_maybe(self)
        if t_self.is_list():
            # scalars or unions are not supported in lists yet (valid graphql spec though)
            return t_self.of_type[0]

    @property
    def is_enum(self) -> Optional[QtGqlEnumDefinition]:
        t_self = optional_maybe(self).type
        with contextlib.suppress(TypeError):
            if issubclass(t_self, AntiForwardRef):
                if isinstance(t_self.resolve(), QtGqlEnumDefinition):
                    return t_self.resolve()

    @property
    def is_builtin_scalar(self) -> Optional[BuiltinScalar]:
        t_self = optional_maybe(self).type
        if isinstance(t_self, BuiltinScalar):
            return t_self

    def is_custom_scalar(self, scalars: CustomScalarMap) -> Optional[Type[BaseCustomScalar]]:
        t_self = optional_maybe(self).type
        if t_self in scalars.values():
            return t_self

    def annotation(self, scalars: CustomScalarMap) -> str:
        """
        :returns: Annotation of the field based on the real type,
        meaning that the private attribute would be of that type.
        this goes for init and the property setter. They are optional by default,
        (at the template) so unwrap optional first
        """
        t_self = optional_maybe(self)

        # int, str, float etc...
        if builtin_scalar := t_self.is_builtin_scalar:
            return builtin_scalar.tp.__name__

        if scalar := t_self.is_custom_scalar(scalars):
            return f"SCALARS.{scalar.__name__}"
        if gql_enum := t_self.is_enum:
            return gql_enum.name
        if model_of := t_self.is_model:
            return f"{QGraphQListModel.__name__}[{model_of.annotation(scalars)}]"
        if object_def := t_self.is_object_type:
            return f"Optional[{object_def.name}]"
        if t_self.is_union():
            return "Union[" + ", ".join(th.annotation(scalars) for th in t_self.of_type) + "]"
        if input_obj := t_self.is_input_object_type:
            return input_obj.name

        raise NotImplementedError  # pragma no cover

    def as_annotation(self, object_map=None):  # pragma: no cover
        raise NotImplementedError("not safe to call on this type")



from __future__ import annotations

from collections import defaultdict
from textwrap import dedent
from typing import TYPE_CHECKING, List, NamedTuple, Optional

import attrs
from graphql import language as gql_lang

from qtgql.codegen.graphql_ref import is_field_node, is_inline_fragment
from qtgql.codegen.py.compiler.template import (
    ConfigContext,
    config_template,
)
from qtgql.codegen.py.objecttype import QtGqlFieldDefinition, QtGqlObjectTypeDefinition
from qtgql.codegen.utils import AntiForwardRef

if TYPE_CHECKING:
    from qtgql.codegen.py.objecttype import QtGqlVariableDefinition


def get_field_from_field_node(
    selection: gql_lang.FieldNode,
    field_type: QtGqlObjectTypeDefinition,
) -> QtGqlQueriedField:
    field_node = is_field_node(selection)
    assert field_node
    return QtGqlQueriedField.from_field(
        field_type.fields_dict[field_node.name.value],
        selection_set=field_node.selection_set,
    )


ID_SELECTION_NODE = (
    gql_lang.FieldNode(name=gql_lang.NameNode(value="id"), arguments=(), directives=()),
)


def inject_id_selection(selection_set: gql_lang.SelectionSetNode) -> None:
    selection_set.selections += ID_SELECTION_NODE


def has_id_field(selection_set: gql_lang.SelectionSetNode) -> bool:
    for field in selection_set.selections:
        assert isinstance(field, gql_lang.FieldNode), f"{field} is not a field"
        if field.name.value == "id":
            return True
    return False


@attrs.define
class QtGqlQueriedField(QtGqlFieldDefinition):
    selections: List[QtGqlQueriedField] = attrs.Factory(list)
    choices: dict[str, List[QtGqlQueriedField]] = attrs.Factory(lambda: defaultdict(list))

    @classmethod
    def from_field(
        cls,
        f: QtGqlFieldDefinition,
        selection_set: Optional[gql_lang.SelectionSetNode],
    ) -> QtGqlQueriedField:
        ret = cls(**attrs.asdict(f, recurse=False))
        if not hasattr(selection_set, "selections"):
            return ret
        assert selection_set
        # inject ID field for types that supports it. unions are handled below.
        if f.can_select_id and not has_id_field(selection_set):
            inject_id_selection(selection_set)

        for selection in selection_set.selections:
            if inline_frag := is_inline_fragment(selection):
                tp = ret.type
                if tp.is_model:  # list of unions.
                    tp = tp.is_model
                assert tp.is_union()  # ATM supported fragments are unions only.
                type_name = inline_frag.type_condition.name.value
                concrete = next(t for t in tp.of_type if t.type.name == type_name)
                assert issubclass(concrete.type, AntiForwardRef)
                concrete = concrete.type.resolve()
                assert isinstance(concrete, QtGqlObjectTypeDefinition)
                if not has_id_field(inline_frag.selection_set) and concrete.has_id_field:
                    inject_id_selection(inline_frag.selection_set)

                for selection_node in inline_frag.selection_set.selections:
                    field_node = is_field_node(selection_node)
                    assert field_node
                    if field_node.name.value == "__typename":
                        continue
                    ret.choices[type_name].append(get_field_from_field_node(field_node, concrete))
            else:
                # unions have inline fragments so this should be either
                f_type = f.type.is_object_type
                if not f_type and f.type.is_model:
                    f_type = f.type.is_model.is_object_type
                assert f_type
                field_node = is_field_node(selection)
                assert field_node
                ret.selections.append(get_field_from_field_node(field_node, f_type))
        return ret

    def as_conf_string(self) -> str:
        return dedent(
            config_template(
                context=ConfigContext(self),
            ),
        )


class QtGqlOperationDefinition(NamedTuple):
    query: str
    name: str
    field: QtGqlQueriedField
    directives: list[str] = []
    fragments: list[str] = []
    variables: list[QtGqlVariableDefinition] = []

    @property
    def operation_config(self) -> str:
        return self.field.as_conf_string()



{% if context.selections or context.choices %}SelectionConfig(
    selections={
        {% for name, selection in context.selections.items() %}
        "{{name}}": {%if selection %} {{selection}} {% else %} None {% endif %},{% endfor %}
    },
{% if context.choices %}
choices={
{% for choice_name, selections in context.choices.items() %}
    "{{choice_name}}": SelectionConfig(selections={
                       {% for field_name, inner_template in selections.items() %} "{{field_name}}": {{inner_template}}, {% endfor %}
}),
{% endfor %}
}
{% endif %}
)
{% endif %}


{% import "macros.jinja.py" as macros %}
from __future__ import annotations
from PySide6.QtCore import Signal, QObject, QEnum

from PySide6.QtQuick import QQuickItem
from typing import Optional, Union
from enum import Enum, auto
from PySide6.QtQml import QmlElement, QmlSingleton

from qtgql.codegen.py.runtime.queryhandler import SelectionConfig, OperationMetaData
from qtgql.tools import qproperty
from qtgql.codegen.py.runtime.bases import QGraphQListModel, NodeRecord, QGraphQLInputObjectABC

{% for dep in context.dependencies %}
{{dep}}{% endfor %}


QML_IMPORT_NAME = "generated.{{context.config.env_name}}.types"
QML_IMPORT_MAJOR_VERSION = 1

__TYPE_MAP__: dict[str, type[{{context.base_object_name}}]] = {}


{% for enum in context.enums %}
class {{enum.name}}(Enum):
    {% for member in enum.members %}
    {{member.name}} = auto()
    """{{member.description}}"""{% endfor %}

{% endfor %}

{% if context.enums %}
@QmlElement
class Enums(QObject):
    {% for enum in context.enums %}
    QEnum({{enum.name}})
    {% endfor %}
{% endif %}


class SCALARS:
    {% for scalar in context.custom_scalars %}
    {{scalar}} = {{scalar}}{% endfor %}

{% for type in context.types %}
class {{ type.name }}({{context.base_object_name}}):
    """{{  type.docstring  }}"""


    def __init__(self, parent: QObject = None, {% for f in type.fields %} {{f.name}}: Optional[{{f.annotation}}] = None, {% endfor %}):
        super().__init__(parent){% for f in type.fields %}
        self.{{  f.private_name  }} = {{f.name}} if {{f.name}} else {{f.default_value}}{% endfor %}
    {%for f in type.fields %}
    {{f.signal_name}} = Signal()

    def {{f.setter_name}}(self, v: {{f.annotation}}) -> None:
        self.{{f.private_name}} = v
        self.{{f.signal_name}}.emit()

    @qproperty(type={{f.property_type}}, fset={{f.setter_name}}, notify={{f.signal_name}})
    def {{f.name}}(self) -> {{f.fget_annotation}}:
        {{f.fget}}
    {% endfor %}
    
    def loose(self, metadata: OperationMetaData) -> None:
        metadata=metadata  {# no-op for types without children #}
        {# loose children #}
        {% for f in type.fields -%}
        {% set private_name %}self.{{f.private_name}}{% endset %}
        {{ macros.loose_field(f, private_name) }}
        {% endfor %}
        {# loose self #}
        {% if type.id_is_optional %}
        if self._id:
            self.__store__.loose(self, metadata.operation_name)
        {% elif type.has_id_field and not type.id_is_optional %}
        self.__store__.loose(self, metadata.operation_name)
        {% else %} {# type with no ID wouldn't clear up itself at the store. delete it here. #}
        self.deleteLater()
        {% endif %}


    @classmethod
    def from_dict(cls, parent, data: dict, config: SelectionConfig, metadata: OperationMetaData) -> {{type.name}}:
        {% if type.id_is_optional %}
        if id_ := data.get('id', None):
            if instance := cls.__store__.get_node(id_):
                instance.update(data, config, metadata)
                return instance
        {% elif type.has_id_field %}
        if instance := cls.__store__.get_node(data['id']):
            instance.update(data, config, metadata)
            return instance
        {% endif %}
        inst = cls(parent=parent)
        {% for f in type.fields -%}
        {% set assign_to %}inst.{{f.private_name}}{% endset %}
        {{ macros.deserialize_field(f,  assign_to) | indent(8)}}
        {%- endfor %}
        {% if type.id_is_optional %}
        if inst.id:
            record = NodeRecord(node=inst, retainers=set()).retain(metadata.operation_name)
            cls.__store__.add_record(record)
        {% elif type.has_id_field and not type.id_is_optional %}
        record = NodeRecord(node=inst, retainers=set()).retain(metadata.operation_name)
        cls.__store__.add_record(record)
        {% endif %}
        return inst

    def update(self, data, config: SelectionConfig, metadata: OperationMetaData) -> None:
        parent = self.parent()
        {%for f in type.fields %}{% set fset %}self.{{f.setter_name}}{% endset %}{% set private_name %}self.{{f.private_name}}{% endset %}
        {{ macros.update_field(f, fset_name=fset, private_name=private_name) | indent(8, True) }}{% endfor %}

__TYPE_MAP__['{{ type.name }}'] = {{ type.name }}
{% endfor %}

{# --------- INPUT OBJECTS ---------- #}

{% for type in context.input_objects %}
class {{type.name}}(QGraphQLInputObjectABC):
    """{{  type.docstring  }}"""


    def __init__(self, parent: QObject = None, {% for f in type.fields %} {{f.name}}: {{f.annotation}} = None, {% endfor %}):
        super().__init__(parent){% for f in type.fields %}
        self.{{f.name}} = {{f.name}}{% endfor %}


    def asdict(self) -> dict:
        ret = {}
        {% for f in type.fields %}{% set attr_name %}self.{{f.name}}{% endset %}
        if {{attr_name}}:
            ret['{{f.name}}'] = {{f.json_repr(attr_name)}}
        {% endfor %}
        return ret


{% endfor %}


from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    NamedTuple,
    Optional,
    TypeVar,
)

from PySide6.QtCore import QObject, Signal
from PySide6.QtQuick import QQuickItem

from qtgql.codegen.py.runtime.environment import get_gql_env
from qtgql.tools import qproperty, slot

if TYPE_CHECKING:
    from qtgql.gqltransport.client import GqlClientMessage

T_QObject = TypeVar("T_QObject", bound=QObject)


class SelectionConfig(NamedTuple):
    selections: Dict[str, Optional[SelectionConfig]] = {}
    choices: Dict[str, SelectionConfig] = {}


class OperationMetaData(NamedTuple):
    operation_name: str
    selections: Optional[SelectionConfig] = None


T = TypeVar("T")


class BaseOperationHandler(QObject, Generic[T]):
    # class-vars
    ENV_NAME: ClassVar[str]
    OPERATION_METADATA: ClassVar[OperationMetaData]
    _message_template: ClassVar[GqlClientMessage]

    # signals
    dataChanged = Signal()
    completedChanged = Signal()
    operationOnFlightChanged = Signal()
    error = Signal(list)

    setVariables: Callable  # implementation by jinja.

    def __init__(self, parent: Optional[QObject]):
        super().__init__(parent)
        name = self.__class__.__name__
        self.setObjectName(name)
        self._completed: bool = False  # whether the graphql operation completed.
        self._operation_on_the_fly: bool = False
        self._data: Optional[T] = None
        self.environment = get_gql_env(self.ENV_NAME)
        self._variables: dict = {}
        self._consumer: Optional[QmlOperationConsumerABC] = None

    def consume(self, consumer: QmlOperationConsumerABC) -> None:
        self._consumer = consumer
        self.dataChanged.connect(consumer.handlerDataChanged)

    def loose(self) -> None:
        """Releases retention from all children, real implementation is
        generated."""
        raise NotImplementedError

    @slot
    def fetch(self) -> None:
        if not self._operation_on_the_fly and not self._completed:
            self.set_operation_on_flight(True)
            self.environment.client.execute(self)  # type: ignore

    @slot
    def refetch(self) -> None:
        self.set_completed(False)
        self.fetch()

    @property
    def message(self) -> GqlClientMessage:
        self._message_template.payload.variables = self._variables
        return self._message_template

    @qproperty(QObject, notify=dataChanged)
    def data(self) -> Optional[T]:
        return self._data

    def set_completed(self, v: bool) -> None:
        self._completed = v
        self.set_operation_on_flight(False)
        self.completedChanged.emit()

    @qproperty(bool, notify=completedChanged)
    def completed(self):
        return self._completed

    def set_operation_on_flight(self, v: bool) -> None:
        self._operation_on_the_fly = v
        self.operationOnFlightChanged.emit()

    @qproperty(bool, notify=operationOnFlightChanged)
    def operationOnFlight(self) -> bool:
        return self._operation_on_the_fly

    def on_data(self, message: dict) -> None:  # pragma: no cover
        # real is on derived class.
        raise NotImplementedError

    def on_completed(self) -> None:
        # https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md#complete
        self.set_completed(True)

    def on_error(self, message: list[dict[str, Any]]) -> None:
        # https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md#error
        self.set_completed(True)
        self.error.emit(message)

    @slot
    def dispose(self) -> None:
        """Get rid of ownership on related data and set data to null."""
        self.loose()

    def deleteLater(self) -> None:
        self.dispose()
        super().deleteLater()


class BaseQueryHandler(BaseOperationHandler[T]):
    """Each handler will be exposed to QML and."""

    ...


class BaseMutationHandler(BaseOperationHandler[T]):
    @slot
    def commit(self) -> None:
        # ATM this is just an alias, we might add some functionality here later like optimistic updates.
        self.refetch()


class QmlOperationConsumerABC(QQuickItem, Generic[T]):
    # signals
    operationNameChanged = Signal()
    autofetchChanged = Signal()
    handlerDataChanged = Signal()

    def __init__(self, parent: Optional[QQuickItem] = None):
        super().__init__(parent)
        self._autofetch: bool = False
        self._handler: BaseOperationHandler[T] = self._get_handler()
        self._handler.consume(self)
        self.destroyed.connect(self._handler.dispose)  # type: ignore

    @qproperty(type=BaseOperationHandler, constant=True)
    def handler(self) -> BaseOperationHandler[T]:
        return self._handler

    def handlerData(self):  # property
        # real implementation in template.
        raise NotImplementedError

    @slot
    def set_autofetch(self, v: bool) -> None:
        self._autofetch = v
        self.autofetchChanged.emit()

    @qproperty(type=bool, notify=autofetchChanged, fset=set_autofetch)
    def autofetch(self) -> bool:
        return self._autofetch

    def componentComplete(self) -> None:
        if self._autofetch:
            self._handler.fetch()
        super().componentComplete()

    @classmethod
    def _get_handler(cls) -> BaseOperationHandler[T]:
        # real implementation is generated.
        raise NotImplementedError



from __future__ import annotations

import warnings
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    TypedDict,
    Union,
)

import graphql
from graphql import OperationType
from graphql import language as gql_lang
from graphql.language import visitor
from graphql.type import definition as gql_def

from qtgql.codegen.graphql_ref import (
    is_enum_definition,
    is_input_definition,
    is_list_definition,
    is_named_type_node,
    is_non_null_definition,
    is_nonnull_node,
    is_object_definition,
    is_operation_def_node,
    is_scalar_definition,
    is_union_definition,
)
from qtgql.codegen.py.compiler.builtin_scalars import BuiltinScalars
from qtgql.codegen.py.compiler.query import QtGqlOperationDefinition, QtGqlQueriedField
from qtgql.codegen.py.compiler.template import (
    TemplateContext,
    handlers_template,
    schema_types_template,
)
from qtgql.codegen.py.objecttype import (
    EnumMap,
    EnumValue,
    GqlTypeHinter,
    QtGqlEnumDefinition,
    QtGqlFieldDefinition,
    QtGqlInputFieldDefinition,
    QtGqlInputObjectTypeDefinition,
    QtGqlObjectTypeDefinition,
    QtGqlVariableDefinition,
)
from qtgql.codegen.utils import anti_forward_ref
from qtgql.exceptions import QtGqlException

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.py.compiler.config import QtGqlConfig

introspection_query = graphql.get_introspection_query(descriptions=True)


class GeneratedNamespace(TypedDict):
    handlers: str
    objecttypes: str


class QtGqlVisitor(visitor.Visitor):
    """Creates handlers for root operations.

    Also injects id field to types that have id field but not explicitly
    queried it.
    """

    def __init__(self, evaluator: SchemaEvaluator):
        super().__init__()
        self.query_handlers: dict[str, QtGqlOperationDefinition] = {}
        self.mutation_handlers: dict[str, QtGqlOperationDefinition] = {}
        self.evaluator = evaluator

    def _get_root_type_field(
        self,
        root_type: QtGqlObjectTypeDefinition,
        gql_field: gql_lang.FieldNode,
    ) -> QtGqlQueriedField:
        return QtGqlQueriedField.from_field(
            root_type.fields_dict[gql_field.name.value],
            gql_field.selection_set,
        )

    def _parse_variable_definition(
        self,
        var: gql_lang.VariableDefinitionNode,
    ) -> QtGqlVariableDefinition:
        return QtGqlVariableDefinition(
            name=var.variable.name.value,
            type=self.evaluator.get_type(var.type),
            type_map=self.evaluator._objecttypes_def_map,
            scalars=self.evaluator.config.custom_scalars,
            enums=self.evaluator._enums_def_map,
            default_value=var.default_value,
        )

    def enter_operation_definition(self, node, key, parent, path, ancestors) -> None:
        if operation := is_operation_def_node(node):
            if operation.operation in (OperationType.QUERY, OperationType.MUTATION):
                root_field: gql_lang.FieldNode = operation.selection_set.selections[0]  # type: ignore
                assert operation.name, "QtGql enforces operations to have names."
                op_name = operation.name.value
                operation_vars: list[QtGqlVariableDefinition] = []
                if variables_def := operation.variable_definitions:
                    for var in variables_def:
                        operation_vars.append(self._parse_variable_definition(var))
                if operation.operation is OperationType.QUERY:
                    assert self.evaluator._query_type
                    root_qtgql_field = self._get_root_type_field(
                        self.evaluator._query_type,
                        root_field,
                    )
                    operation_definition = QtGqlOperationDefinition(
                        query=graphql.print_ast(node),
                        name=op_name,
                        field=root_qtgql_field,
                        directives=node.directives,
                        variables=operation_vars,
                    )
                    self.query_handlers[op_name] = operation_definition
                elif operation.operation is OperationType.MUTATION:
                    assert (
                        self.evaluator._mutation_type
                    ), "You don't have any mutations on your schema"
                    root_qtgql_field = self._get_root_type_field(
                        self.evaluator._mutation_type,
                        root_field,
                    )
                    operation_definition = QtGqlOperationDefinition(
                        query=graphql.print_ast(node),
                        name=op_name,
                        field=root_qtgql_field,
                        directives=node.directives,
                        variables=operation_vars,
                    )
                    self.mutation_handlers[op_name] = operation_definition


class SchemaEvaluator:
    def __init__(self, config: QtGqlConfig):
        self.config = config
        self._objecttypes_def_map: dict[str, QtGqlObjectTypeDefinition] = {}
        self._enums_def_map: EnumMap = {}
        self._input_objects_def_map: dict[str, QtGqlInputObjectTypeDefinition] = {}
        self._query_handlers: dict[str, QtGqlOperationDefinition] = {}
        self._mutation_handlers: dict[str, QtGqlOperationDefinition] = {}
        self._query_type: Optional[QtGqlObjectTypeDefinition] = None
        self._mutation_type: Optional[QtGqlObjectTypeDefinition] = None

    def get_type(self, node: gql_lang.TypeNode) -> GqlTypeHinter:
        if nonnull := is_nonnull_node(node):
            return self._evaluate_field_type(
                graphql.type.GraphQLNonNull(
                    self.schema_definition.get_type(nonnull.type.name.value),  # type: ignore
                ),
            )

        if named_type := is_named_type_node(node):
            gql_concrete = self.schema_definition.get_type(named_type.name.value)
            assert gql_concrete
            return self._evaluate_field_type(gql_concrete)
        raise NotImplementedError(node, "Type is not supported as a variable ATM")

    @cached_property
    def schema_definition(self) -> graphql.GraphQLSchema:
        with (self.config.graphql_dir / "schema.graphql").open() as f:
            return graphql.build_schema(f.read())

    @cached_property
    def root_types(self) -> List[Optional[gql_def.GraphQLObjectType]]:
        return [
            self.schema_definition.get_root_type(OperationType.QUERY),
            self.schema_definition.get_root_type(OperationType.MUTATION),
            self.schema_definition.get_root_type(OperationType.SUBSCRIPTION),
        ]

    def _evaluate_field_type(self, t: gql_def.GraphQLType) -> GqlTypeHinter:
        # even though every type in qtgql has a default constructor,
        # hence there is no "real" non-null values
        # we store it as optional for generating nullable checks only for what's required.
        ret = None
        is_optional = True
        if non_null := is_non_null_definition(t):
            t = non_null.of_type
            is_optional = False

        if list_def := is_list_definition(t):
            ret = GqlTypeHinter(type=list, of_type=(self._evaluate_field_type(list_def.of_type),))
        elif scalar_def := is_scalar_definition(t):
            if builtin_scalar := BuiltinScalars.by_graphql_name(scalar_def.name):
                ret = GqlTypeHinter(type=builtin_scalar)
            else:
                ret = GqlTypeHinter(type=self.config.custom_scalars[scalar_def.name])
        elif enum_def := is_enum_definition(t):
            ret = GqlTypeHinter(
                type=anti_forward_ref(name=enum_def.name, type_map=self._enums_def_map),
            )
        elif obj_def := is_object_definition(t):
            ret = GqlTypeHinter(
                type=anti_forward_ref(name=obj_def.name, type_map=self._objecttypes_def_map),
            )
        elif union_def := is_union_definition(t):
            ret = GqlTypeHinter(
                type=Union,
                of_type=tuple(
                    GqlTypeHinter(
                        type=anti_forward_ref(
                            name=possible.name,
                            type_map=self._objecttypes_def_map,
                        ),
                    )
                    for possible in self.schema_definition.get_possible_types(union_def)
                ),
            )
        elif input_def := is_input_definition(t):
            concrete = self.schema_definition.get_type(input_def.name)
            assert isinstance(concrete, gql_def.GraphQLInputObjectType)
            ret = GqlTypeHinter(type=self._evaluate_input_type(input_def))

        if not ret:  # pragma: no cover
            raise NotImplementedError(f"type {t} not supported yet")

        if is_optional:
            return GqlTypeHinter(type=Optional, of_type=(ret,))
        return ret

    def _evaluate_field(self, name: str, field: gql_def.GraphQLField) -> QtGqlFieldDefinition:
        return QtGqlFieldDefinition(
            type=self._evaluate_field_type(field.type),
            name=name,
            type_map=self._objecttypes_def_map,
            scalars=self.config.custom_scalars,
            enums=self._enums_def_map,
            description=field.description,
        )

    def _evaluate_input_field(
        self,
        name: str,
        field: gql_def.GraphQLInputField,
    ) -> QtGqlInputFieldDefinition:
        return QtGqlInputFieldDefinition(
            type=self._evaluate_field_type(field.type),
            name=name,
            type_map=self._objecttypes_def_map,
            scalars=self.config.custom_scalars,
            enums=self._enums_def_map,
            description=field.description,
        )

    def _evaluate_object_type(
        self,
        type_: gql_def.GraphQLObjectType,
    ) -> Optional[QtGqlObjectTypeDefinition]:
        t_name: str = type_.name
        if evaluated := self._objecttypes_def_map.get(t_name, None):
            return evaluated
        if type_ not in self.root_types:
            try:
                id_field = type_.fields["id"]
                if nonull := is_non_null_definition(id_field.type):
                    id_scalar = graphql.type.scalars.GraphQLID
                    if nonull.of_type is not id_scalar:
                        raise QtGqlException(
                            f"id field type must be of the {id_scalar} scalar!"
                            f"\n Got: {nonull.of_type}",
                        )
                else:
                    warnings.warn(
                        stacklevel=2,
                        message="It is best practice to have id field of type ID!"
                        f"\ntype {type_} has: {id_field}",
                    )
            except KeyError:
                warnings.warn(
                    stacklevel=2,
                    message="QtGql enforces types to have ID field"
                    f"type {type_} does not not define an id field.\n"
                    f"fields: {type_.fields}",
                )
        ret = QtGqlObjectTypeDefinition(
            name=t_name,
            docstring=type_.description,
            fields_dict={
                name: self._evaluate_field(name, field) for name, field in type_.fields.items()
            },
        )
        assert ret not in self.root_types
        self._objecttypes_def_map[ret.name] = ret
        return ret

    def _evaluate_input_type(self, type_: gql_def.GraphQLInputObjectType):
        ret = self._input_objects_def_map.get(type_.name, None)
        if not ret:
            ret = QtGqlInputObjectTypeDefinition(
                name=type_.name,
                docstring=type_.description,
                fields_dict={
                    name: self._evaluate_input_field(name, field)
                    for name, field in type_.fields.items()
                },
            )
            self._input_objects_def_map[ret.name] = ret
        return ret

    def _evaluate_enum(self, enum: gql_def.GraphQLEnumType) -> Optional[QtGqlEnumDefinition]:
        name: str = enum.name

        if self._enums_def_map.get(name, None):
            return None

        return QtGqlEnumDefinition(
            name=name,
            members=[
                EnumValue(name=name, description=val.description or "")
                for name, val in enum.values.items()
            ],
        )

    def parse_schema_concretes(self) -> None:
        for name, type_ in self.schema_definition.type_map.items():
            if name.startswith("__"):
                continue
            if object_definition := is_object_definition(type_):
                if object_type := self._evaluate_object_type(object_definition):
                    if object_definition is self.schema_definition.query_type:
                        self._query_type = object_type
                    elif object_definition is self.schema_definition.mutation_type:
                        self._mutation_type = object_type

            elif enum_def := is_enum_definition(type_):
                if enum := self._evaluate_enum(enum_def):
                    self._enums_def_map[enum.name] = enum

    def parse_operations(self) -> None:
        with (self.config.graphql_dir / "operations.graphql").open() as f:
            operations = graphql.parse(f.read())

        # validate the operation against the static schema
        if errors := graphql.validate(self.schema_definition, operations):
            raise QtGqlException([error.formatted for error in errors])

        # get QtGql fields from the query AST and inject ID field.
        operation_miner = QtGqlVisitor(self)
        visitor.visit(operations, operation_miner)
        self._query_handlers.update(operation_miner.query_handlers)
        self._mutation_handlers.update(operation_miner.mutation_handlers)

    def dumps(self) -> GeneratedNamespace:
        """:return: The generated modules as a string."""
        self.parse_schema_concretes()
        self.parse_operations()
        context = TemplateContext(
            enums=list(self._enums_def_map.values()),
            types=[
                t
                for name, t in self._objecttypes_def_map.items()
                if name not in BuiltinScalars.keys()
            ],
            queries=list(self._query_handlers.values()),
            mutations=list(self._mutation_handlers.values()),
            input_objects=list(self._input_objects_def_map.values()),
            config=self.config,
        )
        return GeneratedNamespace(
            handlers=handlers_template(context),
            objecttypes=schema_types_template(context),
        )

    def dump(self):
        """:param file: Path to the directory the codegen would dump to."""
        for fname, content in self.dumps().items():
            with (self.config.graphql_dir / (fname + ".py")).open("w") as fh:
                fh.write(content)



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, ClassVar, Generic, NamedTuple, Optional, TypeVar

from PySide6.QtCore import QAbstractListModel, QByteArray, QObject, Qt, Signal, Slot

if TYPE_CHECKING:
    from typing_extensions import Self

    from qtgql.codegen.py.runtime.queryhandler import OperationMetaData, SelectionConfig
from qtgql.tools import qproperty, slot

__all__ = ["QGraphQListModel", "get_base_graphql_object"]


class QGraphQLInputObjectABC(QObject):
    def asdict(self) -> dict:
        raise NotImplementedError


class _BaseQGraphQLObject(QObject):
    _id: str
    id: str
    __singleton__: Self
    __store__: ClassVar[QGraphQLObjectStore[Self]]

    def __init_subclass__(cls, **kwargs):
        cls.__store__ = QGraphQLObjectStore()

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)

    @qproperty(str, constant=True)
    def typename(self) -> str:
        return self.__class__.__name__

    @classmethod
    def from_dict(cls, parent: QObject, data: dict, config: SelectionConfig):
        """Creates a new instance from GraphQL raw data."""
        raise NotImplementedError

    def update(self, data, config: SelectionConfig) -> None:
        """updates a node based on new GraphQL data."""
        raise NotImplementedError

    def loose(self, metadata: OperationMetaData) -> None:
        """releases all child objects if exists.

        note that this method would be useful only if the object (or one
        of its children) has an id and a reference in the store,
        otherwise the pointer to this object is release and this object
        would be deleted.
        """
        raise NotImplementedError

    @classmethod
    def default_instance(cls) -> Self:
        # used for default values.
        try:
            return cls.__singleton__  # type: ignore
        except AttributeError:
            cls.__singleton__ = cls()
            return cls.__singleton__


T_BaseQGraphQLObject = TypeVar("T_BaseQGraphQLObject", bound=_BaseQGraphQLObject)


class NodeRecord(NamedTuple):
    node: _BaseQGraphQLObject
    retainers: set[str]  # set of operation names.

    def retain(self, operation_name: str) -> Self:
        self.retainers.add(operation_name)
        return self


class QGraphQLObjectStore(Generic[T_BaseQGraphQLObject]):
    def __init__(self) -> None:
        self._data: dict[str, NodeRecord] = {}

    def get_node(self, id_: str) -> Optional[_BaseQGraphQLObject]:
        assert id_
        if found := self._data.get(id_, None):
            return found.node

    def add_record(self, record: NodeRecord):
        assert record.node.id
        self._data[record.node.id] = record

    def loose(self, node: T_BaseQGraphQLObject, operation_name: str) -> None:
        assert node.id
        with contextlib.suppress(
            KeyError,
        ):  # This node was already deleted, we can safely ignore it
            record = self._data[node.id]
            record.retainers.remove(operation_name)
            if not record.retainers:
                self._data.pop(node.id)
                node.deleteLater()  # we can delete it now since it has no retainers.


class QGraphQListModel(QAbstractListModel, Generic[T_BaseQGraphQLObject]):
    OBJECT_ROLE = Qt.ItemDataRole.UserRole + 1
    _role_names = {OBJECT_ROLE: QByteArray("object")}  # type: ignore
    currentIndexChanged = Signal()

    def __init__(
        self,
        parent: Optional[QObject],
        data: list[T_BaseQGraphQLObject],
    ):
        super().__init__(parent)
        self._data = data
        self._current_index: int = 0

    @slot
    def set_current_index(self, i: int) -> None:
        self._current_index = i
        self.currentIndexChanged.emit()

    @qproperty(int, notify=currentIndexChanged, fset=set_current_index)
    def currentIndex(self) -> int:
        return self._current_index

    @qproperty(QObject, notify=currentIndexChanged)  # type: ignore
    def currentObject(self) -> Optional[T_BaseQGraphQLObject]:
        return self._data[self._current_index]

    def rowCount(self, *args, **kwargs) -> int:
        return len(self._data)

    def roleNames(self) -> dict:
        return self._role_names  # type: ignore

    def data(self, index, role=...) -> Optional[T_BaseQGraphQLObject]:
        if index.row() < len(self._data) and index.isValid():
            if role == self.OBJECT_ROLE:
                return self._data[index.row()]
            raise NotImplementedError(
                f"role {role} is not a valid role for {self.__class__.__name__}",
            )

    def append(self, node: T_BaseQGraphQLObject) -> None:
        count = self.rowCount()
        self.beginInsertRows(self.index(count), count, count)
        self._data.append(node)
        self.endInsertRows()

    @slot
    def pop(self, index: Optional[int] = None) -> None:
        index = -1 if index is None else index
        real_index = index if index > -1 else self.rowCount()
        self.beginRemoveRows(self.index(index - 1).parent(), real_index, real_index)
        self._data.pop(index)
        self.endRemoveRows()

    @slot
    def clear(self) -> None:
        if self._data:
            self.beginRemoveRows(self.index(0).parent(), 0, self.rowCount())
            self._data = []
            self.endRemoveRows()

    @Slot(int, QObject)
    def insert(self, index: int, v: T_BaseQGraphQLObject):
        model_index = self.index(index)
        if index <= self.rowCount() - 1:
            self.beginInsertRows(model_index, index, index)
            self._data[index] = v
            self.endInsertRows()
        else:
            model_index = self.index(self.rowCount() - 1)
            index = self.rowCount()
            self.beginInsertRows(model_index, index, index)
            self._data.append(v)
            self.endInsertRows()

    def update(self, data: list[dict], node_selection: SelectionConfig) -> None:
        raise NotImplementedError

    def removeRows(self, row: int, count: int, parent=None) -> bool:
        # check that this is a valid removal.
        if row + count <= self.rowCount():
            self.beginRemoveRows(self.index(0).parent(), row, count)
            end = self._data[row + count :]
            start = self._data[:row]
            self._data = start + end
            self.endRemoveRows()
            return True
        return False  # pragma: no cover


def get_base_graphql_object(name: str) -> type[_BaseQGraphQLObject]:
    """
    :param name: valid attribute name (used by codegen to import it).
    :returns: A type to be extended by all generated types.
    """
    return type(name, (_BaseQGraphQLObject,), {"__store__": QGraphQLObjectStore()})  # type: ignore


BaseGraphQLObject = get_base_graphql_object("BaseGraphQLObject")

T_BaseModel = TypeVar("T_BaseModel", bound=QGraphQListModel)



{% macro deserialize_field(f, assign_to, include_selection_check=True) -%}
{% if include_selection_check %}
if '{{f.name}}' in config.selections.keys():
{% endif %}
    field_data = data.get('{{f.name}}', {{f.default_value}})
    {%if not (f.type.is_builtin_scalar or f.is_custom_scalar) %}
    {% if include_selection_check -%}
    inner_config = config.selections['{{f.name}}']
    {% else %}
    inner_config = config
    {% endif %}
    {% endif %}
    {% if f.type.is_object_type -%}
    if field_data:
        {{ assign_to }} = {{f.type.is_object_type.name}}.from_dict(
        parent,
        field_data,
        inner_config,
        metadata,
    )
    {% elif f.type.is_model -%}
    {% if f.type.is_model.is_object_type -%}
    {{ assign_to }} = QGraphQListModel(
        parent=parent,
        data=[{{f.type.is_model.is_object_type.name}}.from_dict(parent, data=node, config=inner_config, metadata=metadata) for
              node in field_data],
    )
    {% elif f.type.is_model.is_union -%}
    model_data = []
    for node in field_data:
        type_name = node['__typename']
        choice = inner_config.choices[type_name]
        model_data.append(
            __TYPE_MAP__[type_name].from_dict(self, node,
                                              choice, metadata)
        )
    {{ assign_to }} = QGraphQListModel(parent, data=model_data)
    {% endif %}
    {% elif f.type.is_builtin_scalar -%}
    {{ assign_to }} = field_data
    {% elif f.is_custom_scalar -%}
    {{ assign_to }} = SCALARS.{{f.is_custom_scalar.__name__}}.deserialize(field_data)
    {% elif f.type.is_enum -%}
    {{ assign_to }} = {{f.type.is_enum.name}}[field_data]
    {% elif f.type.is_union() -%}
    type_name = field_data['__typename']
    choice = inner_config.choices[type_name]
    {{ assign_to }} = __TYPE_MAP__[type_name].from_dict(parent, field_data, choice, metadata)
    {% endif %}
{%- endmacro %}



{% macro update_field(f, fset_name, private_name, include_selection_check=True) -%}
{% if include_selection_check %}
if '{{f.name}}' in config.selections.keys():
{% endif %}
    field_data = data.get('{{f.name}}', {{f.default_value}})
    {%if not (f.type.is_builtin_scalar or f.is_custom_scalar) %}
    {% if include_selection_check -%}
    inner_config = config.selections['{{f.name}}']
    {% else %}
    inner_config = config
    {% endif %}
    {% endif %}
    {% if f.type.is_object_type %}
    if not field_data:
        {{fset_name}}(None)
    else:
        {% if f.can_select_id %}
        if {{private_name}} and {{private_name}}._id == field_data.get('id', None):
            {{private_name}}.update(field_data, inner_config, metadata)
        else:
        {% endif %}
            {{fset_name}}({{f.type.is_object_type.name}}.from_dict(
                parent,
                field_data,
                inner_config,
                metadata
            ))
    {% elif f.type.is_model %}
    node_config = inner_config
    new_len = len(field_data)
    prev_len = {{private_name}}.rowCount()
    if new_len < prev_len:
        # crop the list to the arrived data length.
        {{private_name}}.removeRows(new_len, prev_len - new_len)
    for index, node in enumerate(field_data):
        id_ = node.get("id", None)
        if id_ and {{private_name}}._data[index].id == id_:
            # same node on that index just call update there is no need call model signals.
            {{private_name}}._data[index].update(field_data[index], node_config, metadata)
        else:
            # get or create node if wasn't on the correct index.
            # Note: it is safe to call [].insert(50, 50) (although index 50 doesn't exist).
            {% if f.type.is_model.is_object_type %}
            {{private_name}}.insert(index,
                                      {{f.type.is_model.is_object_type.name}}.from_dict(self, field_data[index], node_config, metadata))
            {% elif f.type.is_model.is_union %}
            type_name = node['__typename']
            choice = node_config.choices[type_name]
            {{private_name}}.insert(index,
                                      __TYPE_MAP__[type_name].from_dict(self, field_data[index],
                                                                        choice, metadata))
            {% endif %}
    {% elif f.type.is_builtin_scalar %}
    if {{private_name}} != field_data:
        {{fset_name}}(field_data)
    {% elif f.is_custom_scalar %}
    new = SCALARS.{{f.is_custom_scalar.__name__}}.deserialize(field_data)
    if new != {{private_name}}:
        {{fset_name}}(new)
    {% elif f.type.is_enum %}
    if {{private_name}}.name != field_data:
        {{fset_name}}({{f.type.is_enum.name}}[field_data])
    {% elif f.type.is_union() %}
    type_name = field_data['__typename']
    choice = inner_config.choices[type_name]
    if {{private_name}} and {{private_name}}._id == field_data['id']:
        {{private_name}}.update(field_data, choice, metadata)
    else:
        {{fset_name}}(__TYPE_MAP__[type_name].from_dict(parent, field_data, choice, metadata))
    {% endif %}
{%- endmacro %}


{% macro loose_field(f, private_name) -%}
        {% if f.type.is_object_type or f.type.is_union() %}
        if {{private_name}}:
            {{private_name}}.loose(metadata)
            {{private_name}} = None
        {% elif f.type.is_model.is_object_type or f.type.is_model.is_union %}
        if {{private_name}}:
            for node in {{private_name}}._data:
                node.loose(metadata)
            {{private_name}}.deleteLater()
            {{private_name}} = None
        {% endif %}
{%- endmacro %}



from __future__ import annotations

from typing import Callable, Optional, Type, TypeVar

from graphql import language as gql_lang
from graphql.type import definition as gql_def

T_AST_Node = TypeVar("T_AST_Node", bound=gql_lang.Node)


def ast_identifier_factory(
    expected: Type[T_AST_Node],
) -> Callable[[gql_lang.Node], Optional[T_AST_Node]]:
    def type_guarder(node: gql_lang.ast.Node) -> Optional[T_AST_Node]:
        if isinstance(node, expected):
            return node

    return type_guarder


is_selection_set = ast_identifier_factory(gql_lang.ast.SelectionSetNode)
is_inline_fragment = ast_identifier_factory(gql_lang.InlineFragmentNode)
is_operation_def_node = ast_identifier_factory(gql_def.OperationDefinitionNode)
is_field_node = ast_identifier_factory(gql_def.FieldNode)
is_nonnull_node = ast_identifier_factory(gql_lang.NonNullTypeNode)
is_named_type_node = ast_identifier_factory(gql_lang.NamedTypeNode)


T_Definition = TypeVar("T_Definition", bound=gql_def.GraphQLType)


def definition_identifier_factory(
    expected: Type[T_Definition],
) -> Callable[[gql_def.GraphQLType], Optional[T_Definition]]:
    def type_guarder(definition: gql_def.GraphQLType) -> Optional[T_Definition]:
        if isinstance(definition, expected):
            return definition

    return type_guarder


is_object_definition = definition_identifier_factory(gql_def.GraphQLObjectType)
is_enum_definition = definition_identifier_factory(gql_def.GraphQLEnumType)
is_list_definition = definition_identifier_factory(gql_def.GraphQLList)
is_scalar_definition = definition_identifier_factory(gql_def.GraphQLScalarType)
is_union_definition = definition_identifier_factory(gql_def.GraphQLUnionType)
is_non_null_definition = definition_identifier_factory(gql_def.GraphQLNonNull)
is_input_definition = definition_identifier_factory(gql_def.GraphQLInputObjectType)



from qtgql.tools import slot, qproperty

{% import "macros.jinja.py" as macros %}

from typing import Optional, Union
from PySide6.QtCore import Signal, QObject
from PySide6.QtQml import QmlElement
from qtgql.codegen.py.runtime.queryhandler import BaseQueryHandler, QmlOperationConsumerABC, SelectionConfig, OperationMetaData, BaseMutationHandler
from qtgql.gqltransport.client import  GqlClientMessage, QueryPayload
from qtgql.codegen.py.runtime.bases import QGraphQListModel
from .objecttypes import * # noqa


QML_IMPORT_NAME = "generated.{{context.config.env_name}}"
QML_IMPORT_MAJOR_VERSION = 1


{% macro operation_classvars(operation_def) %}
    ENV_NAME = "{{context.config.env_name}}"
    OPERATION_METADATA = OperationMetaData(
        operation_name="{{operation_def.name}}",
        {% if operation_def.operation_config %}selections={{operation_def.operation_config}}{% endif %}
    )
    _message_template = GqlClientMessage(payload=QueryPayload(query="""{{operation_def.query}}""", operationName="{{operation_def.name}}"))
{% endmacro %}

{% macro operation_common(operation_def) %}
    def set_data(self, d: {{operation_def.field.annotation}}) -> None:
        self._data = d
        self.dataChanged.emit()

    def update(self, data: dict) -> None:
        parent = self
        metadata = self.OPERATION_METADATA
        config = self.OPERATION_METADATA.selections

        {{macros.update_field(operation_def.field, fset_name='self.set_data', private_name='self._data',
                              include_selection_check=False) | indent(4, True)}}

    def loose(self) -> None:
        metadata = self.OPERATION_METADATA
        {% set private_name %}self._data{% endset %}
        {{macros.loose_field(operation_def.field, private_name)}}


    def deserialize(self, data: dict) -> None:
        metadata = self.OPERATION_METADATA
        config = self.OPERATION_METADATA.selections
        parent = self
        {% set assign_to %}self._data{% endset %}
        {{macros.deserialize_field(operation_def.field, assign_to, include_selection_check=False) | indent(4)}}
        self.dataChanged.emit()
    def on_data(self, message: dict) -> None:
        if not self._data:
            self.deserialize(message)

        # data existed and arrived data was null, empty data.
        elif not message.get('{{operation_def.field.name}}', None):
            self._data = None
            self.dataChanged.emit()
        # data existed already, update the data
        else:
            self.update(message)
    {% if operation_def.variables %}
    @slot
    def setVariables(self,
                     {% for var in operation_def.variables %}{{var.name}}: {{var.annotation}}, {% endfor %}
                     ) -> None:
        {% for var in operation_def.variables %}
        if {{var.name}}:
            self._variables['{{var.name}}'] =  {{var.json_repr()}}
        {% endfor %}

    {% endif %}

{% endmacro %}


{% macro operation_consumer_common(operation) %}
    dataChanged = Signal()

    @qproperty(type=QObject, notify=dataChanged)
    def handlerData(self) -> {{operation.field.annotation}}:
        return self._handler._data
{% endmacro %}


{% for query in context.queries %}
class {{query.name}}(BaseQueryHandler[{{query.field.annotation}}]):

    {{operation_classvars(query)}}
    {{operation_common(query)}}

@QmlElement
class Consume{{query.name}}(QmlOperationConsumerABC):
    {{operation_consumer_common(query)}}
    def _get_handler(self) -> BaseQueryHandler[{{query.field.annotation}}]:
        return {{query.name}}(self)
{% endfor %}


{% for mutation in context.mutations %}
class {{mutation.name}}(BaseMutationHandler[{{mutation.field.annotation}}]):

    {{operation_classvars(mutation)}}
    {{operation_common(mutation)}}

@QmlElement
class Consume{{mutation.name}}(QmlOperationConsumerABC[{{mutation.field.annotation}}]):
    {{operation_consumer_common(mutation)}}

    def _get_handler(self) -> BaseMutationHandler[{{mutation.field.annotation}}]:
        return {{mutation.name}}(self)


{% endfor %}

